import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.distributions import Normal
from torch.autograd import Variable

def init_weights(m):
    if isinstance(m, nn.Linear):
        nn.init.normal_(m.weight, mean=0., std=0.1)
        nn.init.constant_(m.bias, 0.1)

class ActorCritic(nn.Module):
    def __init__(self, num_inputs, num_outputs, hidden_size, std=0.0):
        super(ActorCritic, self).__init__()

        #self.base_model = nn.Sequential(
        #    nn.Linear(num_inputs, hidden_size),
        #    nn.Tanh(),
        #    nn.Linear(hidden_size, hidden_size),
        #    nn.Tanh(),
        #)
        self.critic = nn.Sequential(
            nn.Linear(num_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            nn.Linear(hidden_size, 1),
        )
        self.actor = nn.Sequential(
            nn.Linear(num_inputs, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            nn.Linear(hidden_size, hidden_size),
            nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            #nn.Linear(hidden_size, hidden_size),
            #nn.Tanh(),
            nn.Linear(hidden_size, num_outputs),
        )
        self.log_std = nn.Parameter(torch.ones(1, num_outputs) * std)
        self.apply(init_weights)
        # Insert layers for rnn
        
    def forward(self, x):
        #base_value = self.base_model(x)
        value  = self.critic(x)
        mu     = self.actor(x)
        std    = self.log_std.squeeze().exp().expand_as(mu)
        dist   = Normal(mu, std)
        return dist, value, mu


