'''
MAML implementation reference: https://github.com/infinitemugen/MAML-Pytorch

'''
import math
import random
import copy
import gym
from virtual_reachy import ReacherEnv
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
import torch.nn.functional as F
from torch.distributions import Normal
from torch.autograd import Variable
from copy import deepcopy
import time
import matplotlib.pyplot as plt
import argparse
import sys

from captum.attr import IntegratedGradients
from captum.attr import LayerConductance
from captum.attr import NeuronConductance

from actor_critic_small import ActorCritic


class MyLSTM (nn.Module):
    def __init__(self, input_dim, hidden_dim, num_layers, output_dim):
        super(MyLSTM, self).__init__()
        self.rnn = nn.LSTM(input_size = input_dim, hidden_size = hidden_dim, num_layers = num_layers)
        self.fc1 = nn.Linear(hidden_dim, output_dim) #hidden_dim)
        #self.fc2 = nn.Linear(hidden_dim, output_dim)

    def forward(self, x, hidden):
        out, hidden = self.rnn(x, hidden)
        out = self.fc1(out)
        #out = self.fc2(out)
        return out, hidden

class MyLSTM_train(object):
    def __init__(self, input_dim, hidden_dim, num_layers, output_dim, lstm_lr= 5e-3, valid_p = 0.1):
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.num_layers = num_layers
        self.output_dim = output_dim
        self.lr = lstm_lr
        self.valid_p = 0.1
        self.model = MyLSTM(self.input_dim, self.hidden_dim, self.num_layers, self.output_dim)
        self.loss_fn = torch.nn.MSELoss()
        self.optimizer = torch.optim.Adam(self.model.parameters(), lr=self.lr)
        self.losses = []
        self.valid_correct = []
        self.epoch = 1

    def train(self, train_x, train_y, init_hidden):

        loss = 0
        hidden = init_hidden
        out, hidden = self.model(train_x, hidden)
        loss = loss + self.loss_fn(out, train_y)
        self.optimizer.zero_grad()
        loss.backward()
        self.losses.append(loss)
        if (args.lstm_test != 'y'):
            self.optimizer.step()
            #print('loss: ', loss.data.item())

        pred_valid = self.predict(train_x, hidden)
        # Save everything
        state = {
                'model': lstm.model.state_dict(),
                'optimizer':lstm.optimizer.state_dict(),
                }
        torch.save(state, global_lstm_path)
        return loss.data.item(), pred_valid
            

    def predict(self, x, init_hidden):
        test_x = x #self.input_list_to_batch(x)
        hidden = init_hidden
        out, hidden = self.model(test_x, hidden)
        output = out[-1,-1].detach().numpy()
        return output


class PPO(): 
    def __init__ (self, net, optimizer, alpha, widgets, ranges, meta_tasks, num_steps, mini_batch_size, lstm, 
            step_per_epi = 1000, save_every=5, c2 = 0.2, c3=0.05, early_stop=1e-4,ppo_epochs = 4, clip_param=0.2): 
        self.net = net
        self.alpha = alpha
        self.widgets = widgets
        self.num_widgets = len(self.widgets)
        self.num_steps = num_steps
        self.step_per_epi = step_per_epi
        self.ranges = ranges
        self.meta_tasks = meta_tasks
        self.mini_batch_size = mini_batch_size
        self.meta_optimizer = optimizer 
        self.save_every = save_every
        self.ppo_epochs = ppo_epochs
        self.c2 = c2
        self.c3 = c3
        self.early_stop = early_stop
        self.clip_param = clip_param
        self.lstm = lstm

    def compute_gae(self, next_value, rewards, masks, values, gamma=0.9, tau=0.95):
        values = values + [next_value]
        gae = 0
        returns = []
        for step in reversed(range(len(rewards))):
            delta = rewards[step]  + gamma * values[step + 1] * masks[step] - values[step]
            gae = delta + gamma * tau * masks[step] * gae
            returns.insert(0, gae + values[step])
        return returns

    def ppo_iter(self, mini_batch_size, states, actions, log_probs, returns, advantage):
        batch_size = states.size(0)
        for _ in range(batch_size // mini_batch_size):
            rand_ids = np.random.randint(0, batch_size, mini_batch_size)
            yield states[rand_ids, :], actions[rand_ids, :], log_probs[rand_ids, :], returns[rand_ids, :], advantage[rand_ids, :]


    def ppo_update(self, mini_batch_size, states, actions, log_probs, returns, advantages):
        for epoch_count in range(self.ppo_epochs):
            for state, action, old_log_probs, return_, advantage in self.ppo_iter(mini_batch_size, states, actions, log_probs, returns, advantages):
                if (global_v_importance == True):
                    state.requires_grad_()
                    attr, delta = ig.attribute(state,target=1, return_convergence_delta=True)
                    attr = attr.detach().numpy()
                    feature_names = np.linspace(1,num_inputs,num_inputs)
                    feature_names = feature_names.astype("|S4")
                    visualize_importances(feature_names, np.mean(attr, axis=0))
                dist, value, _ = self.net(state)
                entropy = dist.entropy().mean()
                new_log_probs = dist.log_prob(action)

                ratio = (new_log_probs - old_log_probs).exp()
                surr1 = ratio * advantage
                surr2 = torch.clamp(ratio, 1.0 - self.clip_param, 1.0 + self.clip_param) * advantage

                kl_div=np.mean(np.sum(old_log_probs.exp().numpy() * (old_log_probs - new_log_probs).detach().numpy(), axis=1))
                if (kl_div > self.early_stop and epoch_count > 4):
                    print("early stopping because high kl_div at ep: ", epoch_count)
                    return None

                actor_loss  = - torch.min(surr1, surr2).mean()
                critic_loss = (return_ - value).pow(2).mean()

                loss = actor_loss + self.c2 * critic_loss - self.c3 * entropy

                self.meta_optimizer.zero_grad()
                loss.backward()
                self.meta_optimizer.step()

    def sample_position(self):
        sample_pos = []
        sample_pos.append(np.random.uniform(self.ranges[0][0], self.ranges[0][1]))
        sample_pos.append(np.random.uniform(self.ranges[1][0], self.ranges[1][1]))
        return sample_pos

    def sample_widget(self):
        this_widget = 0
        if (if_shuffle):
            random.shuffle(widget_list)
            this_widget = widget_list[0] 
        else:
            this_widget = global_widget
        return this_widget

    def sample_size(self,wid):
        sample_s = []
        if (wid == 5):
            this_size = np.random.uniform(flat_button_size[0][0], flat_button_size[0][1])
            sample_s.append(this_size)
            sample_s.append(this_size)
            sample_s.append(0.001)
        if (wid == 0):
            this_size = np.random.uniform(button_size[0][0], button_size[0][1])
            sample_s.append(this_size)
            sample_s.append(this_size)
            sample_s.append(0.03)
        if (wid == 1):
            sample_s.append(np.random.uniform(slider_size[0][0], slider_size[0][1]))
            sample_s.append(np.random.uniform(slider_size[1][0], slider_size[1][1]))
            sample_s.append(0.04)
        if (wid == 2):
            sample_s.append(np.random.uniform(h_slider_size[0][0], h_slider_size[0][1]))
            sample_s.append(np.random.uniform(h_slider_size[1][0], h_slider_size[1][1]))
            sample_s.append(0.03)
        if (wid == 3):
            this_size = np.random.uniform(doorknob_size[0][0], doorknob_size[0][1])
            sample_s.append(this_size)
            sample_s.append(this_size)
            sample_s.append(np.random.uniform(doorknob_size[1][0], doorknob_size[1][1]))
        if (wid == 6):
            sample_s.append(np.random.uniform(s_slider_size[0][0], s_slider_size[0][1]))
            sample_s.append(np.random.uniform(s_slider_size[1][0], s_slider_size[1][1]))
            sample_s.append(0.04)
        if (wid == 7):
            this_size = np.random.uniform(button_size[0][0], button_size[0][1])
            sample_s.append(0.025)
            sample_s.append(0.05)
            sample_s.append(0.04)
        return sample_s

    def lstm_train_predict(self, this_traj):
        tmp_states = torch.cat(this_traj)
        if (len(tmp_states) < lstm_length):
            for i in range(lstm_length - len(tmp_states)):
                tmp_states = torch.cat([torch.stack([tmp_states[0]]), tmp_states], 0)
        tmp_states = torch.cat([torch.stack([tmp_states[0]]), tmp_states], 0)
        this_x = [tmp_states[-lstm_length: , 0:7]]
        this_x = torch.stack(this_x)
        tmp_y = tmp_states[0, -4:]
        this_y = []
        for i in range(lstm_length):
            this_y.append(tmp_y)
        this_y = torch.stack(this_y)
        this_y = torch.stack([this_y])
        ### TRAINING LSTM HERE
        h0 = Variable(torch.zeros(lstm_num_layers, lstm_length, lstm_hidden_dim).float())
        c0 = Variable(torch.zeros(lstm_num_layers, lstm_length, lstm_hidden_dim).float())
        if (args.lstm_test != 'y'):
            lstm_loss, lstm_pred = self.lstm.train(this_x, this_y, (h0, c0))
        else:
            lstm_pred = self.lstm.predict(this_x, (h0, c0))

        print(lstm_pred)
        lstm_pred = lstm_pred.argmax()
        motion_pred = 0 
        if (lstm_pred == 0):
            motion_pred = 3
            global_pred_task = [1,0,0,0]
        elif (lstm_pred == 1):
            motion_pred = 4
            global_pred_task = [0,1,0,0]
        elif (lstm_pred == 2):
            motion_pred = 7
            global_pred_task = [0,0,1,0]
        elif (lstm_pred == 3):
            motion_pred = 8
            global_pred_task = [0,0,0,1]
        this_correct = 0
        if (motion_pred == sample_wid):
            this_correct = 1
        print('pred correct: ', this_correct)
        global global_lstm_correct_ls
        global_lstm_correct_ls.append(this_correct)
        if(len(global_lstm_correct_ls) == 20):
            print('====================> ave correc rate: ', sum(global_lstm_correct_ls)/len(global_lstm_correct_ls))
            global_lstm_correct_ls = []
        this_traj = []
        print(global_pred_task)


    def outer_loop(self, num_episodes, widget):
        for episode in range(num_episodes):
            print ('a new outer loop')
            outer_reward_sum = 0
            collected_episode_count = 1
            outer_mean = 0

            sample_wid = self.sample_widget()
            sample_pos = self.sample_position() 
            sample_s = self.sample_size(sample_wid)
            print ('a new sample: ', sample_wid, ' at ', sample_pos, 'size: ', sample_s)
            state = envs[sample_wid].reset(sample_pos[0], sample_pos[1], sample_s)

            log_probs = []
            values    = []
            states    = []
            actions   = []
            rewards   = []
            masks     = []
            entropy = 0
            task_reward = 0
            success_count = 0
            this_traj = []

            for step_count in range(self.num_steps):
                state = torch.FloatTensor([state]).to(device)
                dist, value, _ = self.net(state)

                action = dist.sample()
                next_state, reward, done, done2, _ = envs[sample_wid].step(action[0].cpu().numpy())

                log_prob = dist.log_prob(action)
                entropy += dist.entropy().mean()

                log_probs.append(log_prob)
                values.append(value)
                rewards.append(torch.FloatTensor([reward]).unsqueeze(1).to(device))
                done = np.array([done])
                masks.append(torch.FloatTensor(1 - done).unsqueeze(1).to(device))

                states.append(state)
                actions.append(action)

                this_traj.append(state.detach())
                
                state = next_state
                task_reward += reward
                if (global_render):
                    envs[sample_wid].render()
                    time.sleep(0.04)

                if (done and global_lstm):
                    lstm_train_pred(this_traj)
                    
                if done or ((step_count+1)% self.step_per_epi == 0 and step_count != (self.num_steps -1)):
                    sample_wid = self.sample_widget()
                    sample_pos = self.sample_position() 
                    sample_s = self.sample_size(sample_wid)
                    print ('a new sample: ', sample_wid, ' at ', sample_pos, 'size: ', sample_s)
                    state = envs[sample_wid].reset(sample_pos[0], sample_pos[1], sample_s)
                    if (done):
                        success_count += 1
                        if (global_render):
                            time.sleep(0.3)
                            envs[sample_wid].render()
                            time.sleep(0.3)

                    total_rewards.append(task_reward)
                    outer_mean += task_reward
                    task_reward = 0
                    collected_episode_count += 1

                    if(not done):
                        global_pred_task =[0.25, 0.25, 0.25, 0.25]
                

            next_state = torch.FloatTensor([next_state]).to(device)
            _, next_value, _ = self.net(next_state)
            returns = self.compute_gae(next_value, rewards, masks, values)
            returns     = torch.cat(returns).detach()
            log_probs   = torch.cat(log_probs).detach()
            values      = torch.cat(values).detach()
            states      = torch.cat(states)
            actions     = torch.cat(actions)
            advantage   = returns - values
            advantage_n = (advantage - torch.mean(advantage))/ torch.std(advantage)
            
            ### This is making the first gradient descent in the inner loop
            if (args.test != 'y'):
                self.ppo_update(self.mini_batch_size, states, actions, log_probs, returns, advantage_n)

            print ('outer_reward_mean: ', outer_mean/collected_episode_count)
            print ('success rate: ', success_count/collected_episode_count)
            total_success.append(success_count/collected_episode_count)

            if ((episode +1)% self.save_every == 0 and args.test != 'y'):
                np.savetxt(global_save + '.txt', total_rewards)
                np.savetxt(global_save + '_success.txt', total_success)
                actor_critic_state ={
                    'actor_critic': self.net.state_dict(),
                    'optimizer': self.meta_optimizer.state_dict(),
                }
                torch.save(actor_critic_state, global_save)
                print ('everything is saved at %d episode, reward = %d' 
                    % (episode, len(total_rewards)))


def visualize_importances(feature_names, importances, title="Average Feature Importances", plot=True, axis_title="Features"):
    print(title)
    for i in range(len(feature_names)):
        print(feature_names[i], ": ", '%.3f'%(importances[i]))
    x_pos = (np.arange(len(feature_names)))
    if plot:
        plt.figure(figsize=(12,6))
        plt.bar(x_pos, importances, align='center')
        plt.xticks(x_pos, feature_names, wrap=True)
        plt.xlabel(axis_title)
        plt.title(title)
        plt.show()

def compare_models(model_1, model_2):
    models_differ = 0
    for key_item_1, key_item_2 in zip(model_1.state_dict().items(), model_2.state_dict().items()):
        if torch.equal(key_item_1[1], key_item_2[1]):
            pass
        else:
            models_differ += 1
            if (key_item_1[0] == key_item_2[0]):
                print('Mismtach found at', key_item_1[0])
            else:
                raise Exception
    if models_differ == 0:
        print('Models match perfectly!')            


def wrapped_model(inp):
    return ppo_net(inp)[2]

### Begin of the main function
if __name__ == "__main__":

    ######## 
    # Read in arguments
    # --p path: save/load path. Note: the model and the reward should share the same path, ie, ./ppo/ppo (model) and ./ppo/ppo.txt (reward record)
    # --w widget: specify which environment to simulate; default button
    # --l: load the existing history
    # --r: render the simulation
    # --s seed: specify random seed; default = 0
    global_if_scratch = True
    global_render = False
    global_widget = 0
    global_path = ''
    global_save = ''
    global_lstm_path = ''
    global_seed = 0 
    if_shuffle  = True
    global_v_importance = False
    global_lstm = False
    global_pred_task = [0,0,0,0]

    parser = argparse.ArgumentParser(description = 'arg')
    parser.add_argument('--l', dest='if_load', type=str, help='to load a model or not')
    parser.add_argument('--w', dest='widget', type=str, help='to load a model or not')
    parser.add_argument('--r', dest='if_render', type=str, help='to render animation or not')
    parser.add_argument('--p', dest='path', type=str, help='path to the model')
    parser.add_argument('--s', dest='seed', type=int, help='seed number')
    parser.add_argument('--save', dest='save', type=str, help='save path')
    parser.add_argument('--test', dest='test', type=str, help='test')
    parser.add_argument('--vi', dest='v_importance', type=str, help='visualize importance')
    parser.add_argument('--lstm', dest='lstm', type=str, help='if lstm')
    parser.add_argument('--lstm_p', dest='lstm_p', type=str, help='lstm savepath')
    parser.add_argument('--lstm_l', dest='lstm_load', type=str, help='lstm load')
    parser.add_argument('--lstm_test', dest='lstm_test', type=str, help='lstm test')
    args = parser.parse_args()
    if (args.if_load == 'y'):
        global_if_scratch = False
        print ('loading previous model and reward record')
    if (args.if_render == 'y'):
        global_render = True
        print ('rendering is on')
    if (args.path != None):
        global_path = args.path
        global_save = args.path
    else:
        sys.exit ('no path defined. please specify where to save the file by --p path')
    if (args.lstm == 'y'):
        global_lstm = True
    if (args.save != None):
        global_save = args.save
    if (args.lstm_p != None):
        global_lstm_path = args.lstm_p
    if (args.v_importance == 'y'):
        global_v_importance = True

    if (args.widget != None):
        print("hi")
        print(args.widget)
        if (args.widget == 'button_c'):
            global_widget = 0
        elif (args.widget == 'slider_c'):
            global_widget = 1
        elif (args.widget == 'h_slider_c'):
            global_widget = 2
        elif (args.widget == 'doorknob'):
            global_widget = 3
        elif (args.widget == 'flat_button_c'):
            global_widget = 4
        elif (args.widget == 's_slider_c'):
            global_widget = 5 
        elif (args.widget == 'cube'):
            global_widget = 6
        elif (args.widget == 'fake_button'):
            global_widget = 7
        #elif (args.widget == 'doorknob_r'):
        #    global_widget = 9
        #elif (args.widget == 'h_doorknob'):
        #    global_widget = 10 

    if (args.widget != None):
        if_shuffle = False

    # Start environments
    model_name_0 = 'button_c.xml'
    model_name_1 = 'slider_c.xml'
    model_name_2 = 'h_slider_c.xml'
    model_name_3 = 'doorknob.xml'
    model_name_4 = 'flat_button_c.xml'
    model_name_5 = 's_slider_c.xml'
    model_name_6 = 'cube.xml'
    model_name_7 = 'fake_button.xml'
    envs = [ReacherEnv(model_name_0),ReacherEnv(model_name_1),ReacherEnv(model_name_2),
            ReacherEnv(model_name_3),ReacherEnv(model_name_4),ReacherEnv(model_name_5),
            ReacherEnv(model_name_6), ReacherEnv(model_name_7)]
            
    state = envs[0].reset()
    state = envs[1].reset()
    state = envs[2].reset()
    state = envs[3].reset()
    state = envs[4].reset()
    state = envs[5].reset()
    state = envs[6].reset()
    state = envs[7].reset()

    # GPU if any, otherwise cpu
    use_cuda = torch.cuda.is_available()
    device   = torch.device("cuda" if use_cuda else "cpu")
    
    # Random seed
    torch.manual_seed(global_seed)
    np.random.seed(global_seed)

    # Hyper params:
    hidden_size      = 128 
    lr_1             = 8e-5
    n_steps          = 6000
    m_tasks          = 1 
    m_batch_size     = 2000 
    c2_weight        = 0.5
    c3_weight        = 0.008
    ppo_e            = 8 
    early_s          = 1e-4
    clip_p           = 0.1
    total_rewards    = []
    total_success    = []


    n_epochs         = 100 
    widget_list = [ 0, 2 ]
    if (args.widget != None):
        widget_list = [global_widget]

    # Start a new model and load the previous model and reward history if needed
    num_inputs  = len(envs[0].reset_model())
    print ("num_inputs: ", num_inputs)
    num_outputs = envs[0].action_space.shape[0]

    ppo_net = ActorCritic(num_inputs, num_outputs, hidden_size).to(device)
    ig = IntegratedGradients(wrapped_model)
    ppo_optimizer = torch.optim.Adam(ppo_net.parameters(), lr_1)
    if (not global_if_scratch):
        ppo_net_state = torch.load(global_path)
        ppo_net.load_state_dict(ppo_net_state['actor_critic'])
        ppo_optimizer.load_state_dict(ppo_net_state['optimizer'])
        print ('successfully load model')

        if (args.save == None):
            total_rewards = np.loadtxt(global_path + '.txt')
            total_rewards = total_rewards.tolist()
            print('load rewards successfully', len(total_rewards))
            total_success = np.loadtxt(global_path + '_success.txt')
            total_success = total_success.tolist()
            print('load success successfully', len(total_success))


    ranges_xs = [[0.08, 0.13], [-0.3, -0.35]]
    #ranges_xs = [[0.07, 0.14], [-0.29, -0.36]]

    button_size = [[0.03, 0.04]]
    flat_button_size = [[0.01, 0.03]]
    slider_size = [[0.04, 0.045], [0.01, 0.015]]
    s_slider_size = [[0.03, 0.035], [0.02, 0.025]]
    #original
    h_slider_size = [[0.01, 0.02], [0.04, 0.06]]
    #test size
    #h_slider_size = [[0.01, 0.011], [0.01, 0.011]]

    #h_slider_size1 = [[0.005, 0.01], [0.03, 0.05]]
    #h_slider_size2 = [[0.015, 0.02], [0.03, 0.05]]
    doorknob_size = [[0.01, 0.015],[0.1, 0.13]]  # 0.1 - 0.13
    #doorknob_size = [[0.007, 0.009],[0.07, 0.09]]  # 0.1 - 0.13

    # Hyper params for LSTM
    lstm_hidden_dim = 32 
    lstm_num_layers = 6 
    lstm_output_dim = 4
    lstm_input_dim = 7
    lstm_lr = 5e-3
    lstm_length = 15 
    global_lstm_correct_ls = []
    lstm = MyLSTM_train(lstm_input_dim, lstm_hidden_dim, lstm_num_layers, lstm_output_dim, lstm_lr=lstm_lr)
    if (args.lstm_load == 'y'):
        prev_lstm = torch.load(args.lstm_p)
        lstm.model.load_state_dict(prev_lstm['model'])
        lstm.optimizer.load_state_dict(prev_lstm['optimizer'])

    # Init the PPO and start optimization
    ppo = PPO(ppo_net, optimizer = ppo_optimizer, alpha = lr_1,  widgets = envs, ranges = ranges_xs, lstm = lstm, num_steps=n_steps, 
            mini_batch_size = m_batch_size, meta_tasks = m_tasks, ppo_epochs = ppo_e, c2 = c2_weight, c3 = c3_weight, early_stop = early_s, clip_param = clip_p)

    start = time.time()
    ppo.outer_loop(n_epochs, global_widget)
    end = time.time()
    print("time: ", end - start)

