import numpy as np
import utils
import time
import mujoco_env
import math

class ReacherEnv(mujoco_env.MujocoEnv, utils.EzPickle):
    def __init__(self, model_name, step_limit = 'n'):
        utils.EzPickle.__init__(self)
        mujoco_env.MujocoEnv.__init__(self, model_name, 2, step_limit = step_limit)

    #### Define STEP 
    def step(self, a):
        self.do_simulation(a, self.frame_skip)
        self.step_count += 1
        # step
        target_all = self.get_body_com("target_all")
        keycap = self.get_body_com("keycap")
        
        if (self.model_name == 'button.xml' or self.model_name == 'flat_button.xml'  
                or self.model_name == 'button_c.xml' or self.model_name == 'flat_button_c.xml'  ):
            keycap = keycap + [0,0,0.02]
    
        # get reward
        fingertip_left = self.get_body_com("fingertip_left")
        fingertip_right = self.get_body_com("fingertip_right")
        vec = (fingertip_left + fingertip_right)/2 - keycap
        reward_dist = math.exp(-np.linalg.norm(vec))   - 1
        reward_ctrl = (math.exp(-np.square(a).sum()/100) - 1) * 0.5
        #if (self.model_name != 'flat_button_c_fake.xml' and self.model_name != 'cube.xml'):
        #    reward_dist = - np.linalg.norm(vec)
        #if (self.model_name == 'cube.xml'):
        #    d1 = fingertip_right - keycap 
        #    keycap2 = self.get_body_com("keycap2")
        #    d2 = fingertip_left - keycap2
        #    d11 = np.linalg.norm(d1)
        #    d22 = np.linalg.norm(d2)
        #    reward_dist = - (d11+ d22)/2


        reward_finish = 0
        reward_moving = 0
        done = False
        done2 = False

        if (self.model_name == 'button.xml' or self.model_name == 'button_c.xml'):
            if (keycap[2] < 0.275):
                reward_finish = 1
                done = True
        if (self.model_name == 'fake_button.xml'):
            if (keycap[2] < 0.26): # original value = 0.24
                reward_finish = 1
                done = True
            if (target_all[0] - keycap[0] < -0.06): #and target_all[1] - keycap[1] > -0.2):
                reward_finish = -1 
                reward_moving = 0
                done = True
                done2 = True
            #
        if (self.model_name == 'fake_button_2.xml'):
            if (keycap[2] < 0.253):
                reward_finish = 5
                done = True

        if (self.model_name == 'slider.xml' or self.model_name == 'slider_c.xml'):
            reward_moving = math.exp(- (0.05 - keycap[1] +target_all[1])/0.05) - 1 
            if (target_all[1] - keycap[1] < -0.05): #and target_all[1] - keycap[1] > -0.2):
                reward_finish = 1 
                reward_moving = 0
                done = True
        if (self.model_name == 's_slider_c.xml'):
            reward_moving = math.exp(- (0.04 - keycap[1] +target_all[1])/0.04) - 1 
            if (target_all[1] - keycap[1] < -0.05): #and target_all[1] - keycap[1] > -0.2):
                reward_finish = 1 
                reward_moving = 0
                done = True
            #if ( target_all[1] - keycap[1] < 0):
            #    reward_dist += (- (target_all[1] - keycap[1])) * 5 
            #if (target_all[1] - keycap[1] < -0.05): #and target_all[1] - keycap[1] > -0.2):
            #    reward_finish = 1 
            #    done=True
        if (self.model_name ==  'h_slider_c.xml'):
            reward_moving = math.exp(- (0.05 - keycap[0] +target_all[0])/0.05) - 1 
            if (target_all[0] - keycap[0] < -0.05): #and target_all[1] - keycap[1] > -0.2):
                reward_finish = 1 
                reward_moving = 0
                done = True
            #if ( target_all[0] - keycap[0] < 0):
            #    reward_dist += (- (target_all[0] - keycap[0])) * 5 
            #if (target_all[0] - keycap[0] < -0.05): #and target_all[1] - keycap[1] > -0.2):
            #    reward_finish = 100 
        if (self.model_name == 'flat_button.xml' or self.model_name == 'flat_button_c.xml'):
            if (keycap[2] < 0.72):
                reward_finish = 1
                done = True
        if (self.model_name == 'flat_button_c_fake.xml'):
            if (keycap[2] < 0.70):
                reward_finish = -1000
        if (self.model_name == 'cup_c.xml'):
            d1 = np.linalg.norm(fingertip_left - keycap)
            d2 = np.linalg.norm(fingertip_right - keycap)
            if (d1 < 0.05 and d2< 0.05):
                reward_finish = 100
        if (self.model_name == 'cube.xml'):
            d1 = fingertip_right - keycap 
            keycap2 = self.get_body_com("keycap2")
            d2 = fingertip_left - keycap2
            d11 = np.linalg.norm(d1)
            d22 = np.linalg.norm(d2)
            if (d11 < 0.045 and d22 < 0.045):
                reward_finish = 100
        if (self.model_name == 'doorknob.xml'):
            reward_moving = math.exp(- (0.44 - keycap[2])/0.04) - 1 
            if (keycap[2] > 0.44):
                reward_finish = 1
                reward_moving = 0
                done = True
        if (self.model_name == 'doorknob_r.xml'):
            reward_dist += (keycap[2] - 0.3) * 5 
            if (keycap[2] > 0.38):
                reward_finish = 100
        if (self.model_name == 'h_doorknob.xml'):
            reward_moving = math.exp(- (  target_all[1] - keycap[1] )/0.06) - 1 
            if (keycap[1]  >= target_all[1]):
                reward_finish = 1
                reward_moving = 0
                done = True
            #reward_dist += (keycap[1] - (target_all[1] - 0.06)) * 3 
            #if (keycap[1] - target_all[1] > 0.03):
            #    reward_finish = 100

        #print(reward_dist, reward_ctrl, reward_moving)
        reward = reward_dist + reward_ctrl + reward_moving + reward_finish

        if (self.model_name == 'flat_button_c_fake.xml'):
            if (keycap[2] < 0.70):
                done = True
        if (self.model_name == 'cup_c.xml'):
            d1 = np.linalg.norm(fingertip_left - keycap)
            d2 = np.linalg.norm(fingertip_right - keycap)
            if (d1 < 0.05 and d2< 0.05):
                done = True
        if (self.model_name == 'cube.xml'):
            d1 = fingertip_right - keycap 
            keycap2 = self.get_body_com("keycap2")
            d2 = fingertip_left - keycap2
            d11 = np.linalg.norm(d1)
            d22 = np.linalg.norm(d2)
            if (d11 < 0.045 and d22 < 0.045):
                done = True
        if (self.model_name == 'doorknob_r.xml'):
            if (keycap[2] > 0.38):
                done = True

        if (self.step_count > self.max_step and self.step_limit == 'y'):
            done = True
        ob = self._get_obs()
        return ob, reward, done, done2, dict(reward_dist=reward_dist, reward_ctrl=reward_ctrl)

    def viewer_setup(self):
        self.viewer.cam.trackbodyid = 0
        self.viewer.cam.elevation = -30 
        self.viewer.cam.azimuth = 225 
        #self.viewer.cam.lookat[0] += 0.5         # x,y,z offset from the object (works if trackbodyid=-1)
        #self.viewer.cam.lookat[1] += 0.5
        #self.viewer.cam.lookat[2] += 0.5

    def reset_model(self, button_x=0.0, button_y=0.0, if_random='y', size = [0.01, 0.01, 0.01]):
        self.step_count = 0
        qpos = self.np_random.uniform(low=-0.1, high=0.1, size=self.model.nq) + self.init_qpos
        initial_robot_q = [-0.3, 1.57, 2.67, 0, 0, 0, -0.1]
        #pose for picture initial_robot_q = [-0.3, 1.57, 2.57, 0, 0, 0, -0.1]
        """
        while True:
            self.goal = self.np_random.uniform(low=-.2, high=.2, size=2)
            if np.linalg.norm(self.goal) < 0.2:
                break
        """
        if (button_x == 0 and button_y == 0):
            button_x = np.random.uniform(0.08, 0.18)
            button_y = np.random.uniform(-0.2,-0.35)
        initial_target = [0, 0, 0]
        #if (self.model_name == 'button.xml' or self.model_name == 'slider.xml' or self.model_name == 'flat_button.xml' 
        #        or self.model_name == 'button_c.xml' or self.model_name == 'slider_c.xml' or self.model_name == 'flat_button_c.xml'):
        initial_target = [button_x, button_y, 0]
        qpos[0:7] = initial_robot_q 
        #qpos[7:9] = self.goal
        qpos[7:10] = initial_target
        qvel = self.init_qvel + self.np_random.uniform(low=-.005, high=.005, size=self.model.nv)
        for i in range(len(qvel)):
            qvel[i] = 0.0001
        self.set_state(qpos, qvel)
        self.set_size(size)
        return self._get_obs()

    def _get_obs(self):
        env_number = 0
        target_size = self.sim.model.geom_size[self.sim.model._geom_name2id['target2']]
        base_size = self.sim.model.geom_size[self.sim.model._geom_name2id['target_base']]

        if (self.model_name == 'button.xml'):
            env_parameters = [0.3, 0.3, 0.2, 0.5, 0.5] 
        elif (self.model_name == 'button_c.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'fake_button.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'fake_button_2.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'slider.xml'):
            env_parameters = [0.3, 0.1, 0.4, 0.3, 1]
        elif (self.model_name == 'slider_c.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'flat_button.xml'):
            env_parameters = [0.3, 0.3, 0.0001, 0.5, 0.5]
        elif (self.model_name == 'flat_button_c.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'flat_button_c_fake.xml'):
            env_parameters = [0.3, 0.3, 0.0001, 0.5, 0.5]
        elif (self.model_name == 'cup_c.xml'):
            env_parameters = [0.3, 0.3, 0.0001, 0.5, 0.5]
        elif (self.model_name == 'cube.xml'):
            env_parameters = [0.2, 0.2, 0.2, 0.2, 0.2]
        elif (self.model_name == 'h_slider_c.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        elif (self.model_name == 'doorknob.xml'):
            target_size = self.sim.model.geom_size[self.sim.model._geom_name2id['handle_bottom']]
            env_parameters = [target_size[0], target_size[1], target_size[2], 0, 0]
        elif (self.model_name == 'doorknob_r.xml'):
            env_parameters = [0.1, 0.1, 1, 0, 0]
        elif (self.model_name == 'h_doorknob.xml'):
            env_parameters = [0.1, 1.2, 0.1, 0, 0]
        elif (self.model_name == 's_slider_c.xml'):
            env_parameters = [target_size[0], target_size[1], target_size[2], base_size[0], base_size[1]]
        else:
            print('can not find this env setting')


        if (self.model_name == 'button.xml'):
            task_parameters = [1, 0, 0, 0] 
        elif (self.model_name == 'button_c.xml'):
            task_parameters = [1, 0, 0, 0] 
        elif (self.model_name == 'slider.xml'):
            task_parameters = [0, 1, 0, 0] 
        elif (self.model_name == 'slider_c.xml'):
            task_parameters = [0, 1, 0, 0] 
        elif (self.model_name == 's_slider_c.xml'):
            task_parameters = [0, 1, 0, 0] 
        elif (self.model_name == 'flat_button.xml'):
            task_parameters = [1, 0, 0, 0] 
        elif (self.model_name == 'flat_button_c.xml'):
            task_parameters = [1, 0, 0, 0] 
        #elif (self.model_name == 'cube.xml'):
        #    task_parameters = [0, 0, 1, 0, 0, 0] 
        elif (self.model_name == 'h_slider_c.xml'):
            task_parameters = [0, 0, 1, 0] 
        elif (self.model_name == 'doorknob.xml'):
            task_parameters = [0, 0, 0, 1] 
        elif (self.model_name == 'h_doorknob.xml'):
            task_parameters = [0, 0, 0, 0] 
        else:
            task_parameters = [0, 0, 0, 0]

        return np.concatenate([
            self.sim.data.qpos[0:7],#.flat[2:],
            self.sim.data.qvel[0:7],
            (self.get_body_com("fingertip_left") + self.get_body_com("fingertip_right"))/2 - self.get_body_com("keycap"),
            (self.get_body_com("fingertip_left") + self.get_body_com("fingertip_right"))/2, 
            self.get_body_com("target_all"),
            self.get_body_com("keycap"),
            env_parameters,
        #    task_parameters
        ])
